'''Argument Formatters
'''
from . import util
from . import grid


class Argument:
    '''The base-class for all argument formatters.
    If you want to provide a new argument formatter, 
    just override this class and change either the:

     * ``format_arg`` method that is used to format positional and key-value pairs, or 
     * ``format_value`` (which is called in ``format_arg``) to format python values as a string.
       This is often some form of quoted repr or json formatting.

    
    '''
    prefix = suffix = ''

    @classmethod
    def get(cls, key='fire', *a, **kw):
        '''Return an instance of an argument formatter, based on its name.'''
        # key=fire -> FireArgument, key=None -> Argument
        if not key:
            return cls
        if isinstance(key, cls):
            return key
        if isinstance(key, type) and issubclass(key, cls):
            return key(*a, **kw)
        return util.subclass_lookup(cls, suffix='argument').get(key.lower())(*a, **kw)

    def _format_args(self, *a, **kw):
        '''This prepares all arguments and returns them as positional and keyword arguments.

        The keyword arguments contain both the key and value flag, as applicable.
        '''
        if a and isinstance(a[0], grid._BaseGridItem):
            a, kw = (a[0].positional + a[1:], dict(a[0], **kw))
        a = [self.format_arg(v) for v in a]
        kw = {k: self.format_arg(k, v) for k, v in kw.items()}
        return a, kw

    def __call__(self, *a, indent=None, **kw):
        '''Format arguments as a string.
        
        Arguments:
            *a: positional arguments for the CLI.
            indent (int): The indentation (in spaces) for each argument. If
                ``indent`` is ``None`` the arguments are inline, otherwise 
                the arguments are each on their own line.
            **kw: Keyword arguments for the CLI.

        Returns:
            str: The command-line arguments formatted and ready to be outputted
                to a bash command.
        '''
        a2, kw2 = self._format_args(*a, **kw)
        items = (
            [self.prefix]*bool(self.prefix) + 
            a2 + [x for xs in kw2.values() if xs for x in xs] + 
            [self.suffix]*bool(self.suffix))

        indent = '\\\n' + ' '*indent if indent is not None else ''
        return ' '.join((
            f'{indent}{" ".join(x) if isinstance(x, list) else x}' 
            for x in items))

    def format_arg(self, k, v=...):
        '''Format a key-value pair for the command-line.
        
        If it is a positional argument, the value will be in key and the value
        will be ``...``.

        Arguments:
            k: The key (or value, for positional arguments).
            v: The value. If it's a positional argument, it will be ``v=...``
        '''
        return [self.format_value(k) if v is ... else self.format_value(v)]

    def format_value(self, v):
        '''Format a value for the command-line.'''
        return util.shlex_repr(v)



class FireArgument(Argument):
    '''Argument formatting for Python Fire.

    Example: ``python script.py a b --arg1 c --arg2 d``
    
    * Docs: https://google.github.io/python-fire/guide/
    * Github: https://github.com/google/python-fire
    '''
    kw_fmt = '--{key}={value}'

    def format_arg(self, k, v=...):
        if v is ...:
            return [self.format_value(k)]
        return [self.kw_fmt.format(key=k, value=self.format_value(v))]



class ArgparseArgument(Argument):
    '''Argument formatting for Python's builtin argparse.

    Example: ``python script.py a b --arg1 c --arg2 d``
    
    * Docs: https://docs.python.org/3/library/argparse.html
    '''
    short_opt, long_opt = '-', '--'

    def format_arg(self, k, v=...):
        if v is False:
            return [self.format_key(f'no-{k}')]
        key = self.format_key(k)
        if v is True or v is ...:
            return [key]
        if v is None:
            return []
        if not isinstance(v, (list, tuple, set)):
            v = [v]
        # this will all be joined to gether as strings. The extra
        # list on the outside means that it will appear on the same line
        # when indentation is requested
        vs = [self.format_value(x) for x in v]
        return [ [key] + vs ]

    def format_key(self, k):
        return '{}{}'.format(self.long_opt if len(k) > 1 else self.short_opt, k)


class SacredArgument(FireArgument):
    '''Formatting for sacred.

    Example: ``python script.py with arg1=a arg2=b``
    
    * Docs: https://sacred.readthedocs.io/en/stable/
    * Github: https://github.com/IDSIA/sacred
    '''
    prefix = 'with'
    kw_fmt = '{key}={value}'


class HydraArgument(FireArgument):
    '''Formatting for hydra.

    Example: ``python script.py  db.user=root db.pass=1234``
    
    * Docs: https://hydra.cc/docs/intro/
    * Github: https://github.com/facebookresearch/hydra
    '''
    kw_fmt = '{key}={value}'
    available_prefixes = ['+', '++', '~', '']
    # https://hydra.cc/docs/advanced/override_grammar/basic

    def __init__(self, default_prefix='+', **kw):
        default_prefix = default_prefix or ''
        if default_prefix not in self.available_prefixes:
            raise ValueError('invalid hydra prefix: {}. Must be one of {}'.format(
                default_prefix, self.available_prefixes))
        self.kw_fmt_prefixed = default_prefix + self.kw_fmt
        super().__init__(**kw)

    def format_arg(self, k, v=...):
        if v is ...:
            return [self.format_value(k)]
        kw_fmt = (
            self.kw_fmt
            if any(k.startswith(pfx) for pfx in self.available_prefixes if pfx)
            else self.kw_fmt_prefixed)
        return [kw_fmt.format(key=k, value=self.format_value(v))]



'''Automating Slurm job generation.

Generate a set of `.sbatch` files over a grid of parameters to be searched over. 
A run script is created which will submit all generated jobs as once.

You can also use `ShellBatch` which excludes the slurm/module references so you can 
test & run on your local machine or test server.

> NOTE: because NYU switched to Greene which now utilizes singularity, the previous 
workflow regarding loading modules is not as necessary. Therefore, I will likely be 
experimenting and changing how things are done to suit the new workflow. If you want 
to keep the old setup, pin to `<=0.1.2`. (Sorry I should have just did a `0.2` bump 
off the bat).



'''
import os
import time
import pprint
import jinja2
import pathtree
from .grid import *
from . import util
from .args import Argument

loader = jinja2.ChoiceLoader([
    jinja2.PackageLoader('slurmjobs', 'templates')
])
env = jinja2.Environment(loader=loader)
env.filters['prettyjson'] = util.prettyjson
env.filters['prefixlines'] = util.prefixlines
env.filters['pprint'] = pprint.pformat
env.filters['comment'] = lambda x, ns=1, ch='#', nc=1: util.prefixlines(x, ch*nc+' '*ns)


class Jobs:
    '''The base class for Job generation. Sub-class this if you want to provide your own

    .. code-block:: python

        import slurmjobs

        batch = slurmjobs.SlurmBatch(
            'python train.py',
            email='mynetid@nyu.edu',
            conda_env='my_env')

        # generate jobs across parameter grid
        run_script, job_paths = batch.generate([
            ('kernel_size', [2, 3, 5]),
            ('nb_stacks', [1, 2]),
            ('lr', [1e-4, 1e-3]),
        ], receptive_field=6)


    We use Jinja2 for our script templating. See https://jinja.palletsprojects.com/en/3.0.x/templates/ 
    for documentation about its syntax and structure.
    
    Attributes:
        options (dict): Template options. Anything here is made available to the template. 
        template (str): The Jinja2 template for the sbatch script. 
        job_template (str): The Jinja2 template for the jobs script that launches the sbatch files.
        cli (str, slurmjobs.args.Argument): The command line argument format your script uses.
            By default, this uses Fire. See ``args.FireArgument``.

        job_id_arg (str): The argument name to use for passing the job ID. 
            Set to None to ignore omit the job ID.
        job_id_key_sep (str): The separator between key and value used in the job ID.
        job_id_item_sep (str): The separator between key-value items in the job ID. 
        
        allowed_job_id_chars (str, list): Characters (other than alphanumeric) allowed in the job ID.
            Other characters are removed.
        special_character_replacement (str): A string that can be used in place of 
            special characters not found in ``allowed_job_id_chars``. Default is empty.
        key_abbreviations (dict): A mapping from full name to abbreviation for keys in the job ID.
        abbreviate_length (int): The length to abbreviate keys in the job ID.
        float_precision (int): The number of decimals to limit float values in the job ID.



    '''
    options = {
        'email': None,
        'conda_env': None,
        'run_dir': None,
        'bashrc': True,
        'set_e': False,
    }
    cli = 'fire'
    root_dir = 'jobs'

    # make script templating clearer and more simple
    template = '''{% extends 'job.base.j2' %}
    '''
    run_template = '''{% extends 'run.base.j2' %}
    '''

    job_id_arg = 'job_id'
    job_id_key_sep = '-'
    job_id_item_sep = ','
    allowed_job_id_chars = ',._-'
    special_character_replacement = ''
    omit_missing_keys_from_job_id = True
    
    key_abbreviations = {}
    abbreviate_length = None
    float_precision = None

    def __init__(self, command, name=None, cli=None, 
                 root_dir=None, backup=True, job_id=True, 
                 template=None, run_template=None,
                 **options):
        self.template = template or self.template
        self.run_template = run_template or self.run_template

        # raise an error if any unrecognized arguments were passed.
        wrong_options = set(options) - set(self.options)
        if wrong_options:
            raise TypeError(
                f"Unrecognized options: {wrong_options}. If you extended your template to use additional options, "
                "please give them default values in your Class.options dictionary.")
        
        # copy the options dict and add in new values
        self.options = dict(self.options)
        for k, v in options.items():
            dft = self.options[k]
            if isinstance(dft, dict):
                v = dict(dft, **(v or {}))
            self.options[k] = v

        self.command = command
        self.name = name or util.command_to_name(command)
        self.cli = Argument.get(self.cli if cli is None else cli)
        self.job_id_arg = self.job_id_arg if job_id is True else job_id or None

        # paths
        self.root_dir = root_dir or self.root_dir
        self.paths = self.get_paths()
        if 'batch_dir' in self.paths and len(self.paths.batch_dir.join('*').glob()):
            if backup:
                util.maybe_backup(self.paths.batch_dir)
            else:
                self.paths.batch_dir.rmglob(include=True)

    def format_id_item(self, k, v):
        '''Formats a key-value pair for the job ID. 

        You can override this to change how each key-value pair in a 
        job ID is formatted.

        Arguments:
            k (str): The argument key.
            value (any): The argument value.

        Returns:
            Return a tuple if you want it to be joined using ``job_id_key_sep``. Return a string
            to be used as is. Return None to omit it from the job ID. 

        Be warned that omitting key/value pairs runs the risk of filename collisions which will mean 
        multiple jobs overwritting each other in the same file.
        '''
        # key formatting
        if k in self.key_abbreviations:
            k = self.key_abbreviations[k]
        elif self.abbreviate_length:
            k = k[:self.abbreviate_length]
        # TODO do before abbreviate_length
        k = ''.join(
            c if c in self.allowed_job_id_chars or c.isalnum() else self.special_character_replacement
            for c in k)

        # value formatting
        if self.float_precision is not None and isinstance(v, float):
            v = f'{v:.{self.float_precision}g}'
        return k, v

    def format_job_id(self, args, keys=None, name=None, ignore_keys=None):
        '''Convert a dictionary to a job ID.

        Arguments:
            args (GridItem): The job dictionary.
            keys (list, tuple): The keys that we want to include in the job ID.
            name (str): The job name to include in the job ID. By default
                this will be the name associated with the Jobs object. Pass 
                ``False`` to use no name.
        '''
        name = name or self.name if name is not False else False
        # format each key
        parts = []
        keys = getattr(args, 'grid_keys', args) if keys is None else keys
        if ignore_keys:
            keys = (k for k in keys if k not in ignore_keys)
        for k in keys:
            if k not in args:
                if self.omit_missing_keys_from_job_id:
                    continue
            formatted = self.format_id_item(k, args.get(k))
            if isinstance(formatted, (list, tuple)):
                parts.append(self.job_id_key_sep.join(map(str, formatted)))
            elif formatted is not None:
                parts.append(str(formatted))

        # join them together
        return self.job_id_item_sep.join(map(
            str, [name]*bool(name) + parts)).replace(' ', '')


    def generate(self, grid_=None, *a, ignore_job_id_keys=None, **kw):
        '''Generate slurm jobs for every combination of parameters.
        
        Arguments:
            grid_ (Grid, list): The parameter grid.
            *a: Positional arguments to add to the command.
            **kw: Additional keyword arguments to pass to the command.

        Returns:
            tuple:

                str: The path to the run script.

                list[str]: The list of paths for each job file.
        '''
        # generate jobs
        # kw = dict(kw, **(kwargs_ or {})) # for taken keys.
        grid = Grid.as_grid([{}] if grid_ is None else grid_)

        used = set()
        job_paths = []
        for d in grid:
            d.positional += a
            d.update(kw)
            job_id = self.format_job_id(
                d, d.grid_keys, name=self.name, 
                ignore_keys=ignore_job_id_keys)
            if job_id in used:
                raise RuntimeError(f"Duplicate job ID: {job_id}")
            job_paths.append(self.generate_job(job_id, _grid=grid, _args=d))

        # generate run file
        run_script = self.generate_run_script(job_paths, _grid=grid)

        # store the current timestamp
        if 'time_generated' in self.paths.paths:
            self.paths.time_generated.write(time.time())

        return run_script, job_paths
    
    def generate_job(self, job_id, *a, _args=None, _grid=None, **params):
        '''Generate a single slurm job file'''
        # build command
        paths = self.paths.specify(job_id=job_id)
        if self.job_id_arg:
            params[self.job_id_arg] = job_id
        
        args = _args
        if args is None:
            args = GridItem()
        args.update(params)
        args.positional += a

        # generate job file
        paths.job.up().make()
        with open(paths.job, "w") as f:
            f.write(env.from_string(self.template).render(
                job_id=job_id,
                command=self.command,
                paths=paths,
                args=args,
                cli=self.cli,
                grid=_grid,
                **self.options,
            ).lstrip())

        if 'output' in paths.paths:
            paths.output.up().make()

        return paths.job.format()


    def generate_run_script(self, _job_paths, _grid=None, **kw):
        '''Generate a job run script that will submit all jobs.'''
        # Generate run script
        file_path = self.paths.run
        with open(file_path, "w") as f:
            f.write(env.from_string(self.run_template).render(
                name=self.name,
                command=self.command,
                job_paths=_job_paths,
                paths=self.paths,
                cli=self.cli,
                grid=_grid,
                **self.options, **kw))
        util.make_executable(file_path)
        return file_path

    def get_paths(self, **kw):
        paths = pathtree.tree(self.root_dir, {'{name}': {
            '': 'batch_dir',
            '{job_id}.job.sh': 'job',
            'run.sh': 'run',
            # optional
            'output/{job_id}.log': 'output',
            'time_generated': 'time_generated',
        }}).update(name=self.name, **kw)
        return paths



class Shell(Jobs):
    '''Generate batch jobs to be run without a task scheduler.
    This will create a grid of scripts that will be run using ``nohup``,
    a program that will keep running your script even if the parent 
    process (e.g. your ssh session) dies.

    This is mainly just for doing small tests and whatnot.

    .. code-block:: python

        import slurmjobs

        jobs = Shell('python myscript.py')
        jobs.generate([
            ('a', [1, 2]),
            ('b', [1, 2]),
        ])

    '''

    template = '''{% extends 'job.shell.j2' %}
    '''




class Slurm(Jobs):
    '''Generate jobs for sbatch. This was used for HPC Prince. 
    For HPC Greene, please use ``Singularity``.

    This is functionally equivalent to:

    .. code-block:: python

        # sbatch_args
        #SBATCH --nodes=...
        #SBATCH --cpus-per-task=...
        #SBATCH --gres=gpu:...
        #SBATCH --mail-user=...
        ...

        {# shell_code_body #}
    
    '''
    # make sbatch args more isomorphic
    # while also providing helpers for setting cpus to the number of gpus by default (for example)
    options = dict(
        Jobs.options,
        # sbatch arguments
        # see: https://slurm.schedmd.com/sbatch.html
        sbatch={
            'time': '3-0',
            'mem': '48GB',
            # 'n_gpus': None,
            # 'n_cpus': None,
            # 'nodes': None,
        },
        # n_gpus=0,
        # n_cpus=None,
        # nodes=None,
        modules=None,  # no need for modules with singularity
        bashrc=False,  # disable bashrc by default
    )
    template = '''{% extends 'job.sbatch.j2' %}
    '''
    run_template = '''{% extends 'run.sbatch.j2' %}
    '''

    module_presets = {
        'cuda9': ['cudnn/9.0v7.3.0.29', 'cuda/9.0.176'],
        'cuda10': ['cuda/10.0.130', 'cudnn/10.0v7.4.2.24'],
        'cuda10.1': ['cuda/10.1.105', 'cudnn/10.1v7.6.5.32'],
        'cuda10.2': ['cuda/10.2.89'],
        'cuda11': ['cuda/11.0.194'],
        'cuda11.1': ['cuda/11.1.74'],
        'cuda11.3': ['cuda/11.3.1'],
        'conda': ['anaconda3/2020.07'],
        'anaconda': ['anaconda3/2020.07'],
        'anaconda3': ['anaconda3/2020.07'],
    }

    def __init__(self, *a, sbatch=None, modules=None, n_gpus=None, n_cpus=None, **kw):
        modules = util.flatten(
            self.module_presets.get(m, m) for m in (modules or ()))
        super().__init__(*a, modules=modules, sbatch=sbatch, **kw)

        # handle n_cpus n_gpus
        sbatch = self.options['sbatch']
        n_cpus = sbatch.pop('n_cpus', None)
        n_gpus = sbatch.pop('n_gpus', None)
        gres = sbatch.get('gres')
        if n_cpus:
            if sbatch.get('cpus-per-task'):
                raise ValueError("n_cpus specified as well as cpus-per-task")
            sbatch['cpus-per-task'] = n_gpus or 1 if n_cpus is None else n_cpus
        if n_gpus:
            if sbatch.get('gres'):
                raise ValueError("n_gpus specified as well as gres. Please choose one or the other.")
            sbatch['gres'] = f'gpu:{n_gpus or 0}'
            # sbatch['gpus-per-task'] = n_gpus or 0
        if gres and isinstance(gres, int):
            sbatch['gres'] = f'gpu:{gres}'

    def get_paths(self, **kw):
        paths = pathtree.tree(self.root_dir, {'{name}': {
            '': 'batch_dir',
            '{job_id}.sbatch': 'job',
            'run.sh': 'run',
            'slurm/slurm_%j__{job_id}.log': 'output',
            'time_generated': 'time_generated',
        }}).update(name=self.name, **kw)
        return paths


class Singularity(Slurm):
    '''Generate jobs for sbatch + singularity. Use this for HPC Greene.
    
    This is functionally equivalent to:

    .. code-block:: python

        {# sbatch_args #}

        # execute it inside the container
        singularity exec --overlay ... ... /bin/bash << EOF

        . /ext3/env.sh

        {# shell_code_body #}

        EOF

    To setup a singularity overlay with a anaconda python environment, 
    please feel free to use https://gist.github.com/beasteers/84cf1eb2e5cc7bc4cb2429ef2fe5209e
    I'm in the process of integrating it with ``slurmjobs``.

    Source Tutorial: https://sites.google.com/a/nyu.edu/nyu-hpc/services/Training-and-Workshops/tutorials/singularity-on-greene

    
    '''
    sif_dir = '/scratch/work/public/singularity/'
    options = dict(
        Slurm.options,
        sif='cuda11.0-cudnn8-devel-ubuntu18.04.sif',
        overlay='overlay-5GB-200K.ext3',
        readonly=True,
        overlays=None,
        readonly_overlays=None,
        singularity_init_script='source /ext3/env.sh',
    )
    template = '''{% extends 'job.singularity.j2' %}
    '''
    def __init__(self, command, overlay=None, sif=None, *a, **kw):
        if overlay is not None:
            kw['overlay'] = overlay
        if sif is not None:
            kw['sif'] = sif
        super().__init__(command, *a, **kw)
        self.options['sif'] = self.options['sif'] and os.path.join(
            self.sif_dir, self.options['sif'])



import os
import json
import shlex
import types
import itertools
import collections



def summary(run_script, job_paths):
    '''Print out a nice summary of the job sbatch files and the run script.'''
    print('Generated', len(job_paths), 'job scripts:')
    for p in job_paths:
        print('\t', p)
    print()

    print('To submit all jobs, run:')
    print('.', run_script)
    print()


'''

Misc / template utils

'''

def command_to_name(command):
    '''Convert a bash command into a usable job name.
    
    NOTE: This was originally to turn 'python my/script.py' into 'my.script'
          but obviously this is just a special case and I never really found
          a nice, general way to do this.
    '''
    fbase = os.path.splitext(shlex.split(command)[1])[0]
    return fbase.replace('/', '.').lstrip('.').replace(' ', '-')


def as_chunks(lst, n=1):
    return [lst[i:i + n] for i in range(0, len(lst), n)]



def make_executable(file_path):
    '''Grant permission to execute a file.'''
    # https://stackoverflow.com/a/30463972
    mode = os.stat(file_path).st_mode
    mode |= (mode & 0o444) >> 2
    os.chmod(file_path, mode)


def maybe_backup(path, prefix='~'):
    '''Backup a path if it exists.'''
    # create backup
    if path.exists():
        bkp_previous_path = path.prefix(prefix).next_unique(1)
        os.rename(path, bkp_previous_path)
        print('moved existing', path, 'to', bkp_previous_path)



def prettyjson(value):
    '''Pretty-print data using json.'''
    return json.dumps(value, sort_keys=True, indent=4) if value else ''


def prefixlines(text, prefix='# '):
    '''Prefix each line. Can be used to comment a block of text.'''
    return ''.join(prefix + l for l in str(text).splitlines(keepends=True))


def dict_merge(*ds, depth=-1, **kw):
    '''Recursive dict merge.

    Arguments:
        *ds (dicts): dicts to be merged.
        depth (int): the max depth to be merged.
        **kw: extra keys merged on top of the final dict.

    Returns:
        merged (dict).
    '''
    def merge(dicta, dictb, depth=-1):
        '''Recursive dict merge.

        Arguments:
            dicta (dict): dict to be merged into.
            dictb (dict): merged into dicta.
        '''
        for k in dictb:
            if (depth != 0 and k in dicta
                    and isinstance(dicta[k], dict)
                    and isinstance(dictb[k], collections.Mapping)):
                dict_merge(dicta[k], dictb[k], depth=depth - 1)
            else:
                dicta[k] = dictb[k]

    mdict = {}
    for d in ds + (kw,):
        merge(mdict, d, depth=depth)
    return mdict


def flatten(args, cls=(list, tuple, set, types.GeneratorType)):
    '''Flatten iterables into a flat generator.'''
    if isinstance(args, cls):
        yield from (a for arg in args for a in flatten(arg))
    elif args is not None:
        yield args


def clsattrdiff(cls, base_cls=None, funcs=False):
    '''Get the non-function attributes that exist on ``cls``, but not on ``base_cls``.'''
    attrs = dict()
    base_dict = base_cls.__dict__ if base_cls is not None else {}
    i = cls.__mro__.index(base_cls) if base_cls is not None else None
    for cls_i in cls.__mro__[:i][::-1]:
        for k, v in cls_i.__dict__.items():
            if (not k.startswith('_') and 
                    k not in base_dict and 
                    (funcs or not callable(v))):
                attrs[k] = v
    return attrs


'''

Argument

'''


def all_subclasses(cls):
    '''Return all recursive subclasses of a class.'''
    return set(cls.__subclasses__()).union(
        s for c in cls.__subclasses__() for s in all_subclasses(c))


def subclass_lookup(cls, prefix='', suffix='', name_override=None):
    '''Get a dictionary of name to subclass (removing any common prefix/suffix).'''
    return {
        (getattr(c, name_override, None) if name_override else None) 
            or stripstr(c.__name__.lower(), prefix, suffix): c
        for c in all_subclasses(cls)
        if not c.__name__.startswith('_')}


def shlex_repr(v):
    '''Prepare a variable for bash. This will serialize the object using json 
    and then quote the variable if it contains any spaces/bash-specific characters.
    '''
    # v = repr(v)
    v = json.dumps(v)
    # if v is a quoted string, remove the quotes
    if len(v) >= 2 and v[0] == v[-1] and v[0] in '"\'':
        v = v[1:-1]
    return shlex.quote(v) # only quote if necessary (has spaces or bash chars)


def escape(txt, chars='"\''):
    '''Escape certain characters from a string. By default, it's quotes.'''
    for c in chars:
        txt = txt.replace(c, '\\' + c)
    return txt


def stripstr(text, prefix='', suffix=''):
    '''Strip a prefix and/or suffix from a string.'''
    text = text[len(prefix):] if prefix and text.startswith(prefix) else text
    text = text[:-len(suffix)] if suffix and text.endswith(suffix) else text
    return text


def get_available_tensorflow_cuda_versions():
    '''Load the tensorflow-cuda lookup table.'''
    # https://www.tensorflow.org/install/source#gpu
    import csv
    with open(os.path.join(os.path.dirname(__file__), 'cuda_versions.csv'), 'r') as f:
        rows = list(csv.DictReader(f, delimiter='\t'))
    versions = ({
        'cuda': d['CUDA'], 'cudnn': d['cuDNN'], 
        'python': d["Python version"].split(', '),
        'version': d['Version'].split('-')[-1]
    } for d in rows)
    return {d['version']: d for d in versions}


def find_tensorflow_cuda_version(version, latest=False):
    '''Given a tensorflow version, lookup the correct'''
    import fnmatch
    versions = get_available_tensorflow_cuda_versions()
    version = str(version).split('.')[:2]
    version = '.'.join(version + ['*']*(3 - len(version)))
    matched = sorted(v for v in versions if fnmatch.fnmatch(v, version))
    if not matched:
        raise ValueError('No value matching {!r}'.format(version))
    return versions[matched[-1 if latest else 0]]

if __name__ == '__main__':
    import fire
    fire.Fire()


'''Batch generate slurm/shell jobs.

Examples:
$ # this will generate slurm sbatch files
$ python -m slurmjobs slurm --cmd='python train.py' generate "{kernel_size: [2,3,5], lr: [1e-4, 1e-3]}"
$ # this will generate shell files that run commands in the background using nohup
$ python -m slurmjobs sh --cmd='python train.py' generate "{kernel_size: [2,3,5], lr: [1e-4, 1e-3]}"

NOTE: Python Fire requires that class __init__ args use keyword notation (--cmd=MY_CMD)

Fire User Guide: https://github.com/google/python-fire/blob/master/docs/guide.md
'''
from slurmjobs import *

if __name__ == '__main__':
    import fire
    fire.Fire({
        'sh': Shell,
        'slurm': Slurm,
    })



'''Parameter grids!

This lets you do basic grid expansion and grid arithmetic.

.. code-block:: python

    g = Grid([
        ('a', [1, 2]),
        ('b', [1, 2]),
    ], name='train')

    # append two configurations
    g = g + LiteralGrid([{'a': 5, 'b': 5}, {'a': 10, 'b': 10}])

    # create a bigger grid from the product of another grid
    g = g * Grid([
        ('c', [5, 6])
    ], name='dataset')

    assert list(g) == [
        {'a': 1, 'b': 1, 'c': 5},
        {'a': 1, 'b': 1, 'c': 6},
        {'a': 1, 'b': 2, 'c': 5},
        {'a': 1, 'b': 2, 'c': 6},
        {'a': 2, 'b': 1, 'c': 5},
        {'a': 2, 'b': 1, 'c': 6},
        {'a': 2, 'b': 2, 'c': 5},
        {'a': 2, 'b': 2, 'c': 6},
        {'a': 5, 'b': 5, 'c': 5},
        {'a': 5, 'b': 5, 'c': 6},
        {'a': 10, 'b': 10, 'c': 5},
        {'a': 10, 'b': 10, 'c': 6},
    ]


'''
import itertools
import collections

# Parameter Grids


class BaseGrid:
    '''The base class for all grids. Use this if you want to extend another grid.
    
    You just need to implement:

     * ``__iter__``: This should yield all items generated by the grid
     * ``__len__``: This should tell you the number of items in the grid. If you cannot
       determine the length of a grid, then raise a TypeError (as the other grids do).
     * ``__repr__``: A nice string representation of the grid
    '''
    grid = ()

    def __init__(self, name=None, ignore_job_id_keys=None, **constants) -> None:
        self.name = name
        self.ignore_job_id_keys = ignore_job_id_keys or []
        self.constants = constants

    def __repr__(self):
        '''A nice string representation of the grid.'''
        return '{}({})'.format(self.__class__.__name__, ', '.join(map(repr(self.grid))))

    def __str__(self):
        ''''''  # NOTE: not sure if this should be different or if we should switch str/repr
        return repr(self)

    def __len__(self):
        '''Get the number of iterations in the grid.
        
        Note that any use of generators or functions without a length will 
        cause this to raise a TypeError.
        '''
        raise NotImplemented

    def __iter__(self):
        '''Yield all combinations from the parameter grid.'''
        raise NotImplemented

    def __add__(self, other):
        '''Combine two parameter grids sequentially.'''
        return GridChain(self, other)

    def __mul__(self, other):
        '''Create a grid as the combination of two grids.'''
        return GridCombo(self, other)

    @classmethod
    def as_grid(cls, grid):
        '''Ensure that a value is a grid.'''
        if isinstance(grid, BaseGrid):
            return grid
        if isinstance(grid, (list, tuple)):
            if all(isinstance(g, dict) for g in grid):
                return LiteralGrid(grid)
        return Grid(grid)


class Grid(BaseGrid):
    '''A parameter grid! To get all combinations from the grid, just do ``list(Grid(...))``.

    Arguments:
        grid (list, dict): The parameter grid. Should be either a dict
            or a list of key values, where the values are a list of values
            to use in the grid. Examples of valid inputs:

            .. code-block:: python
                
                # simple grid
                Grid([ ('a', [1, 2]), ('b', [1, 2]) ])
                Grid({ 'a': [1, 2], 'b': [1, 2] })

                # paired parameters
                Grid([
                    ('a', [1, 2]), 
                    (('b', 'c'), ([1, 2], [1, 2])) 
                ])
                Grid([ 
                    ('a', [1, 2]),
                    [{'b': 1, 'c': 1}, {'b': 2, 'c': 2}],
                ])

                # any of these are valid grid specs
                g = slurmjobs.Grid([
                    # basic
                    ('a', [1, 2]),
                    # paired
                    (('b', 'c'), ([1, 1, 2, 2], [1, 2, 1, 2])),

                    # literal list of dicts
                    [{'d': i} for i in [1, 2]],
                    # dict generator
                    ({'e': i} for i in [1, 2]),
                    # function
                    lambda: [{'f': i} for i in [1, 2]],
                    # function that returns a generator
                    lambda: ({'g': i} for i in [1, 2]),

                    # basic generator
                    ('h', (x for x in [1, 2])),
                    # basic function
                    ('i', lambda: [x for x in [1, 2]]),
                ])
                keys = 'abcdefghi'
                assert list(g) == [
                    dict(zip(keys, vals)) for vals in
                    itertools.product(*([ [1, 2] ]*len(keys)))
                ]

        name (str): The name of this grid. Can be used to search 
            for the parameters from this grid.
        **constants: Extra parameters to add to the grid that don't vary.
            These will not be included in the job_id name.

    .. .. code-block:: python

    ..     g = Grid([
    ..         ('a', [1, 2]),
    ..         ('b', [1, 2]),
    ..     ])
    ..     assert list(g) == [
    ..         {'a': 1, 'b': 1},
    ..         {'a': 2, 'b': 1},
    ..         {'a': 1, 'b': 2},
    ..         {'a': 2, 'b': 2},
    ..     ]

    .. You can also do pairwise parameter expansion.

    .. .. code-block:: python

    ..     g = Grid([
    ..         ('a', [1, 2]),
    ..         (('b', 'c'), ([1, 2], [3, 4])),
    ..     ])
    ..     assert list(g) == [
    ..         {'a': 1, 'b': 1, 'c': 3},
    ..         {'a': 2, 'b': 1, 'c': 3},
    ..         {'a': 1, 'b': 2, 'c': 4},
    ..         {'a': 2, 'b': 2, 'c': 4},
    ..     ]

    Just a heads up, there is nothing stopping you from passing an infinite generator,
    meaning that you can make some fancy sampling grid generators, but ``slurmjobs`` will
    take that and not know when to stop. If you want to use an infinite generator, just 
    wrap it in ``itertools.islice`` which will let you provide a limit.

    Obviously, ``slurmjobs`` doesn't operate anywhere near the memory scale where you'd need to 
    even use generators in the first place, but I figured why limit the implementation if 
    it can be used for other things too.
    '''
    _is_expandable = False
    _easy_access_nested = False
    def __init__(self, __grid, name=None, **constants):
        self.grid = list(__grid.items()) if isinstance(__grid, dict) else __grid
        super().__init__(name, **constants)

    def __repr__(self):
        return '[\n{}]'.format(''.join(map('  {!r},\n'.format, self.grid)))

    def __len__(self):
        return prod(self._as_grid_length(g) for g in self.grid)

    def __getitem__(self, index):
        '''Get the series for a variable name.'''
        res = self._get_by_key(self.grid, index)
        if res is not None:
            return res
        raise KeyError(index)

    def __setitem__(self, index, value):
        res = self._set_by_key(self.grid, index, value)
        if res:
            return
        raise KeyError(index)

    def _get_by_key(self, grid, name):
        for xs in grid:
            try:
                len(xs)
                k, v = xs
            except (TypeError, ValueError):
                continue
            if isinstance(k, dict): continue
            if self._easy_access_nested and not isinstance(name, (list, tuple)) and isinstance(k, (list, tuple)):
                res = self._get_by_key(zip(*xs), name)
                if res is None:
                    continue
                return res
            if k == name:
                return v

    def _set_by_key(self, grid, name, value):
        for i, xs in enumerate(grid):
            try:
                len(xs)  # don't unpack generators
                k, v = xs                
            except (TypeError, ValueError):
                continue
            if isinstance(k, dict): continue
            if self._easy_access_nested and not isinstance(name, (list, tuple)) and isinstance(k, (list, tuple)):
                rep = list(zip(k, v))
                if self._set_by_key(rep, name, value):
                    grid[i] = list(zip(*rep))
                    return True
                continue
            if k == name:
                grid[i] = k, value
                return True
        if self.is_expandable:
            grid.append((name, value))
            return True
        return False

    def _as_grid_length(self, xs):
        '''Determine the length of a grid product item.'''
        try:
            l = len(xs)
            if l == 2 and not isinstance(xs[0], dict):
                if isinstance(xs[0], (list, tuple)):
                    return max(len(x) for x in xs[1])
                return len(xs[1])
            return l
        except (TypeError, IndexError):
            raise TypeError("Could not determine accurate length from grid item: {}".format(repr(xs)))
        

    def _as_grid_product_iter(self, xs):
        '''This returns one iteration to use in the product.
        '''
        if callable(xs):
            xs = xs()
        # checking the first value
        try:
            first = xs[0]  # list of dicts?
        except TypeError:
            # generator of dicts: ({'a': ...} for _ in ...)
            firsts, xs = peek(xs)
            if not firsts:
                return []
            first = firsts[0]
        except IndexError:  # empty list
            return []
        # [{'a': 1, 'b': 1}, {'a': 2, 'b': 2}]
        if isinstance(first, dict):
            return xs
        
        # asserting it's length 2
        key, values = xs
        # if any of the values are functions, call them
        values = [v() if callable(v) else v for v in values]
        # (('a', 'b'), ([1, 2, 3], [3, 4, 5]))
        if isinstance(key, (list, tuple)):
            return (
                collections.OrderedDict([(k, v) for k, v in zip(key, vs) if v != ...])
                for vs in itertools.zip_longest(*values, fillvalue=...))
        # ('a', [1, 2, 3])
        return ({key: v} for v in values)

    def __iter__(self):
        grid = [self._as_grid_product_iter(xs) for xs in self.grid]
        for ds in itertools.product(*grid):
            # expand grid pairs [('a', 'b'), ([1, 2, 3], [1, 2, 3])]
            yield GridItem(
                dict({k: v for d in ds for k, v in d.items()}, **self.constants), 
                [k for d in ds for k in d], self.name, 
                ignore_keys=self.ignore_job_id_keys)


class LiteralGrid(BaseGrid):
    '''A parameter grid, specified as a flattened list. This
    doesn't do any grid expansion, it lets you specify the grid
    as you want.

    Arguments:
        grid (list, dict): The parameter list. Should be a list of 
            dicts, each corresponding to a parameter config.
        name (str): The name of this grid. Can be used to search 
            for the parameters from this grid.

    .. code-block:: python

        g = LiteralGrid([
            {'a': 1, 'b': 1},
            {'a': 1, 'b': 2},
            {'a': 2, 'b': 2},
        ])
        assert list(g) == [
            {'a': 1, 'b': 1},
            {'a': 1, 'b': 2},
            {'a': 2, 'b': 2},
        ]
    '''
    def __init__(self, __grid, name=None, **constants):
        self.grid = [__grid] if isinstance(__grid, dict) else __grid
        super().__init__(name, **constants)

    def __repr__(self):
        return '[\n{}]'.format(''.join(map('  {!r},\n'.format, self.grid)))

    def __len__(self):
        return len(self.grid)

    def __iter__(self):
        for d in self.grid:
            # expand grid pairs [('a', 'b'), ([1, 2, 3], [1, 2, 3])]
            keys = d.grid_keys if isinstance(d, _BaseGridItem) else list(d)
            yield GridItem(
                dict(d, **self.constants), keys, self.name, 
                ignore_keys=self.ignore_job_id_keys)
            # should we just use dict's internal ordering for the keys?



class _BaseGridItem(dict):
    positional = ()
    # def variant_items(self):
    #     return [(k, self[k]) for k in self.grid_keys]


class GridItem(_BaseGridItem):
    '''Represents a dictionary of arguments, the keys that vary, 
    and a name for the group of args.
    '''
    def __init__(self, grid=None, keys=(), name=None, positional=(), ignore_keys=None):
        if ignore_keys:
            keys = [k for k in keys if k not in ignore_keys]
        self.grid_keys = keys
        self.name = name
        self.positional = positional or ()
        super().__init__(() if grid is None else grid)

    def __getitem__(self, key):
        return super().__getitem__(key)

    def find(self, name):
        return self if self.name == name else None

    

class GridItemBundle(_BaseGridItem):
    '''Merges GridItems/GridBundles. Merges the dict, keys, and any groups.
    Can use ``.find(name)`` so search for a subset of items.
    '''
    def __init__(self, *grids, name=None, ignore_keys=None):
        merged = {}
        groups = {}
        keys = []
        for d in grids:
            merged.update(d)
            keys.extend(d.grid_keys or ())
            for ki, di in getattr(d, 'groups', {}).items():
                groups[ki] = dict(groups.get(ki, ()), **di)
            k = d.name
            if k is not None:
                groups[k] = dict(groups.get(k, ()), **d)
        if name is not None:
            groups[name] = dict(groups.get(name, ()), **merged)

        self.name = name
        if ignore_keys:
            keys = (k for k in keys if k not in ignore_keys)
        self.grid_keys = unique(keys)
        self.groups = groups
        super().__init__(merged)

    def __getitem__(self, key):
        if key in self.groups:
            return self.groups[key]
        return super().__getitem__(key)

    def __getattr__(self, key):
        return self.__getitem__(key)

    def find(self, name):
        return self.groups.get(name)



class GridChain(BaseGrid):
    '''This handles the addition of two grids (one after the other).

    You can create this doing ``grid_a + grid_b``. The only reason to 
    use this directly is if you want to give it a name.
    
    .. code-block:: python

        a = Grid(('a', [1, 2]), ('b', [1, 2]))
        b = Grid(('c', [1, 2]))

        # functionally equivalent
        c = a + b
        c = GridChain(a, b, name='my-a-then-b-grid')
        c_items = list(a) + list(b)
    '''
    def __init__(self, *grids, name=None):
        self.grid = grids
        super().__init__(name)

    def __repr__(self):
        return ' + '.join(map(repr, self.grid))

    def __len__(self):
        return sum(len(g) for g in self.grid)

    def __iter__(self):
        for g in self.grid:
            yield from g


class GridCombo(BaseGrid):
    '''This handles the multiplication of two grids (combinations). It 
    will create a grid as a product of all provided grids.

    You can create this doing ``grid_a * grid_b``. The only reason to 
    use this directly is if you want to give it a name or if you want to
    make a grid product of 3 or more grids.
    
    .. code-block:: python

        a = Grid(('a', [1, 2]), ('b', [1, 2]))
        b = Grid(('c', [1, 2]))

        # functionally equivalent
        c = a * b
        c = GridCombo(a, b, name='my-a-b-combo-grid')
        c_items = [
            dict(da, **db)
            for da, db in itertools.product(a, b)
        ]
    '''
    def __init__(self, *grids, name=None):
        self.grid = grids
        super().__init__(name)

    def __repr__(self):
        return ' * '.join(map('({!r})'.format, self.grid))
        
    def __len__(self):
        return prod(len(g) for g in self.grid)

    def __iter__(self):
        for gs in itertools.product(*self.grid):
            yield GridItemBundle(*gs, name=self.name)


class GridOmission(BaseGrid):
    '''This handles the subtraction of two grids (combinations). It will 
    yield only dicts from ``grid_a`` that don't appear in ``grid_b``.

    You can create this doing ``grid_a - grid_b``. The only reason to 
    use this directly is if you want to give it a name.
    
    .. code-block:: python

        a = Grid(('a', [1, 2]), ('b', [1, 2]))
        b = Grid(('a', [2]), ('b', [1]))

        # functionally equivalent
        c = a - b
        c = GridOmission(a, b, name='my-a-minus-b-grid')
        omit = list(b)
        c_items = [da for da in a if da not in omit]
    '''
    def __init__(self, grid, omission, name=None):
        self.grid = grid
        self.omission = omission
        super().__init__(name)

    def __repr__(self):
        return ' - '.join(map('({!r})'.format, [self.grid, self.omission]))
        
    def __len__(self):
        return sum(1 for d in self)

    def __iter__(self):
        omit = list(self.omission)
        for d in self.grid:
            if d not in omit:
                yield d





def prod(ns):
    '''Like ``sum()`` but for products.'''
    total = 1
    for n in ns:
        total *= n
        if total == 0:
            break
    return total


def unique(xs):
    used = set()
    return [x for x in xs if not (x in used or used.add(x))]


def peek(it, n=1):
    it = iter(it)
    first = [x for i, x in zip(range(n), it)]
    return first, (x for xs in (first, it) for x in xs)


import glob
import setuptools

import os, imp
version = imp.load_source(
    'slurmjobs.__version__', 
    os.path.join(os.path.dirname(__file__), 'slurmjobs/__version__.py')).__version__


setuptools.setup(name='slurmjobs',
                 version=version,
                 description='Generate slurm jobs in batches.',
                 long_description=open('README.md').read().strip(),
                 long_description_content_type='text/markdown',
                 author='Bea Steers',
                 author_email='bea.steers@gmail.com',
                 # url='http://path-to-my-packagename',
                 package_data={'slurmjobs': ['templates/*.j2']},
                 scripts=glob.glob('scripts/**/*.sh'),
                 packages=setuptools.find_packages(),
                 install_requires=['path-tree', 'Jinja2'],
                 license=open('README.md').readline().strip(),
                 extras_require={
                    'test': ['pytest', 'pytest-cov'],
                    'doc': ['sphinx']
                },
                 keywords='slurm sbatch job batch generation parameters ml machine learning python')
